import os
import sys
import datetime

HERE = os.path.abspath(os.path.dirname(__file__))
SRC = os.path.normpath(os.path.join(HERE, "..", "src"))
sys.path.insert(0, SRC)

import odin.datetimeutil

ARE_YOU_EXPERIENCED = datetime.date(1967, 5, 12)
MWT = odin.datetimeutil.FixedTimezone(-6, "Mountain War Time")
BOOM = datetime.datetime(1945, 7, 16, 5, 29, 45, 0, MWT)



# -*- coding: utf-8 -*-
from typing import Union, Sequence, Dict, Optional, Type, Any

from six import StringIO
from odin import Resource
from odin import resources, ResourceAdapter
from odin.exceptions import CodecDecodeError
from odin.utils import getmeta

try:
    import toml
except ImportError:  # pragma: no cover
    raise ImportError(  # pragma: no cover
        "odin.codecs.toml_codec requires the 'toml' package."
    )


TOML_TYPES = {}
CONTENT_TYPE = "application/toml"


def load(fp, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load a resource from a TOML encoded file.

    If a ``resource`` value is supplied it is used as the base resource for the supplied YAML. I one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param fp: a file pointer to read TOML data format.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """

    try:
        data = toml.load(fp)
    except toml.TomlDecodeError as ex:
        raise CodecDecodeError(str(ex))
    return resources.build_object_graph(
        data, resource, full_clean, False, default_to_not_supplied,
    )


def loads(s, resource=None, full_clean=True, default_to_not_supplied=False):
    """
    Load a resource from a TOML encoded string.

    If a ``resource`` value is supplied it is used as the base resource for the supplied YAML. I one is not supplied a
    resource type field ``$`` is used to obtain the type represented by the dictionary. A ``ValidationError`` will be
    raised if either of these values are supplied and not compatible. It is valid for a type to be supplied in the file
    to be a child object from within the inheritance tree.

    :param s: a string containing TOML.
    :param resource: A resource type, resource name or list of resources and names to use as the base for creating a
        resource. If a list is supplied the first item will be used if a resource type is not supplied.
    :param full_clean: Do a full clean of the object as part of the loading process.
    :param default_to_not_supplied: Used for loading partial resources. Any fields not supplied are replaced with
        NOT_SUPPLIED.
    :returns: A resource object or object graph of resources loaded from file.

    """

    try:
        data = toml.loads(s)
    except toml.TomlDecodeError as ex:
        raise CodecDecodeError(str(ex))
    return resources.build_object_graph(
        data, resource, full_clean, False, default_to_not_supplied,
    )


class OdinEncoder(toml.TomlEncoder):
    def __init__(
        self, include_virtual_fields=True, include_type_field=True, *args, **kwargs
    ):
        super(OdinEncoder, self).__init__(*args, **kwargs)
        self.include_virtual_fields = include_virtual_fields
        self.include_type_field = include_type_field

    def resource_to_dict(self, v):
        resource_dict = v.to_dict(self.include_virtual_fields)
        if self.include_type_field:
            meta = getmeta(v)
            resource_dict[meta.type_field] = meta.resource_name
        return resource_dict

    def dump_value(self, v):
        if isinstance(v, (Resource, ResourceAdapter)):
            resource_dict = self.resource_to_dict(v)
            return self.dump_inline_table(resource_dict)
        if type(v) in TOML_TYPES:
            v = TOML_TYPES[type(v)](v)
        return super(OdinEncoder, self).dump_value(v)


RT = Union[Resource, ResourceAdapter, Sequence[Resource], Sequence[ResourceAdapter], Dict]


def dump(resource, fp, encoder=None, include_virtual_fields=True, **kwargs):
    # type: (RT, StringIO, Optional[Type[OdinEncoder]], bool, Any) -> None
    """
    Dump to a TOML encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param fp: The file pointer that represents the output file.
    :param encoder: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param include_virtual_fields: Include virtual fields in the output 
    :param kwargs: Additional keyword arguments for the encoder.

    """
    encoder = (encoder or OdinEncoder)(
        include_virtual_fields, resource.__class__, **kwargs
    )

    if isinstance(resource, (Resource, ResourceAdapter)):
        resource = encoder.resource_to_dict(resource)

    toml.dump(resource, fp, encoder)


def dumps(resource, encoder=None, include_virtual_fields=True, **kwargs):
    # type: (RT, Optional[Type[OdinEncoder]], bool, Any) -> str
    """
    Dump to a TOML encoded file.

    :param resource: The root resource to dump to a JSON encoded file.
    :param encoder: Encoder to use serializing to a string; default is the :py:class:`OdinEncoder`.
    :param include_virtual_fields: Include virtual fields in the output 
    :param kwargs: Additional keyword arguments for the encoder.
    :returns: TOML encoded string.

    """
    encoder = (encoder or OdinEncoder)(
        include_virtual_fields, resource.__class__, **kwargs
    )

    if isinstance(resource, (Resource, ResourceAdapter)):
        resource = encoder.resource_to_dict(resource)

    return toml.dumps(resource, encoder)



# -*- coding: utf-8 -*-
try:
    import arrow
except ImportError:
    raise ImportError("The arrow package is not installed.")

try:
    from odin.codecs.json_codec import JSON_TYPES
except ImportError:
    pass
else:
    JSON_TYPES[arrow.Arrow] = str

try:
    from odin.codecs.yaml_codec import YAML_TYPES
except ImportError:
    pass
else:
    YAML_TYPES[arrow.Arrow] = str

try:
    from odin.codecs.msgpack_codec import TYPE_SERIALIZERS
except ImportError:
    pass
else:
    TYPE_SERIALIZERS[arrow.Arrow] = str



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import enum
import pytest

from odin.exceptions import ValidationError
from odin.fields.future import EnumField


class Colour(enum.Enum):
    Red = "red"
    Green = "green"
    Blue = "blue"


class Options(enum.Enum):
    Undefined = ""
    Yes = "True"
    No = "False"


class Number(enum.IntEnum):
    Thirteen = 13
    FortyTwo = 42
    SixtyNine = 69


class TestEnumField(object):
    @pytest.mark.parametrize(
        ("field", "value", "actual"),
        (
            (EnumField(Colour), "red", Colour.Red),
            (EnumField(Colour), "green", Colour.Green),
            (EnumField(Colour), Colour.Blue, Colour.Blue),
            (EnumField(Colour, null=True), None, None),
            (EnumField(Colour, null=True), "", None),
            (EnumField(Options), "", Options.Undefined),
            (EnumField(Number), 13, Number.Thirteen),
            (EnumField(Number), 42, Number.FortyTwo),
            (EnumField(Number), Number.SixtyNine, Number.SixtyNine),
            (EnumField(Number, null=True), None, None),
        ),
    )
    def test_to_python__valid(self, field, value, actual):
        assert field.clean(value) == actual

    @pytest.mark.parametrize(
        ("field", "value"),
        (
            (EnumField(Colour), "pink"),
            (EnumField(Colour), True),
            (EnumField(Colour), None),
            (EnumField(Colour), ""),
            (EnumField(Colour), Number.Thirteen),
            (EnumField(Number), "13"),
            (EnumField(Number), 14),
            (EnumField(Number), None),
        ),
    )
    def test_to_python__invalid(self, field, value):
        with pytest.raises(ValidationError):
            field.clean(value)

    @pytest.mark.parametrize(
        "field, value, actual",
        (
            (EnumField(Colour), "red", None),
            (EnumField(Colour), Colour.Blue, "blue"),
            (EnumField(Colour), None, None),
            (EnumField(Number), 13, None),
            (EnumField(Number), Number.FortyTwo, 42),
            (EnumField(Number), None, None),
        ),
    )
    def test_prepare(self, field, value, actual):
        assert field.prepare(value) == actual

    @pytest.mark.parametrize(
        "field, expected",
        (
            (
                EnumField(Colour),
                ((Colour.Red, "Red"), (Colour.Green, "Green"), (Colour.Blue, "Blue")),
            ),
            (
                EnumField(Number),
                (
                    (Number.Thirteen, "Thirteen"),
                    (Number.FortyTwo, "FortyTwo"),
                    (Number.SixtyNine, "SixtyNine"),
                ),
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                ((Colour.Red, "Red"), (Colour.Blue, "Blue")),
            ),
        ),
    )
    @pytest.mark.skipif("sys.version_info.major < 3")
    def test_choices(self, field, expected):
        assert field.choices == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (
                EnumField(Colour),
                {(Colour.Red, "Red"), (Colour.Green, "Green"), (Colour.Blue, "Blue")},
            ),
            (
                EnumField(Number),
                {
                    (Number.Thirteen, "Thirteen"),
                    (Number.FortyTwo, "FortyTwo"),
                    (Number.SixtyNine, "SixtyNine"),
                },
            ),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                {(Colour.Red, "Red"), (Colour.Blue, "Blue")},
            ),
        ),
    )
    def test_choices__with_set(self, field, expected):
        """Use sets to handle random ordering in earlier Python releases"""
        assert set(field.choices) == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (EnumField(Colour), (("red", "Red"), ("green", "Green"), ("blue", "Blue"))),
            (EnumField(Number), ((13, "Thirteen"), (42, "FortyTwo"), (69, "SixtyNine"))),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                (("red", "Red"), ("blue", "Blue")),
            ),
        ),
    )
    @pytest.mark.skipif("sys.version_info.major < 3")
    def test_choices_doc_text(self, field, expected):
        assert field.choices_doc_text == expected

    @pytest.mark.parametrize(
        "field, expected",
        (
            (EnumField(Colour), {("red", "Red"), ("green", "Green"), ("blue", "Blue")}),
            (EnumField(Number), {(13, "Thirteen"), (42, "FortyTwo"), (69, "SixtyNine")}),
            (
                EnumField(Colour, choices=(Colour.Red, Colour.Blue)),
                {("red", "Red"), ("blue", "Blue")},
            ),
        ),
    )
    def test_choices_doc_text__with_sets(self, field, expected):
        """Use sets to handle random ordering in earlier Python releases"""
        assert set(field.choices_doc_text) == expected



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import copy
import datetime
import six
import uuid

from typing import Sequence, Tuple

from odin import exceptions, datetimeutil, registration
from odin.utils import getmeta, lazy_property
from odin.validators import (
    EMPTY_VALUES,
    MaxLengthValidator,
    MinValueValidator,
    MaxValueValidator,
    validate_url,
    validate_ipv4_address,
    validate_ipv6_address,
    validate_ipv46_address,
    validate_email_address,
)
from .base import BaseField

__all__ = (
    "BooleanField",
    "StringField",
    "UrlField",
    "IntegerField",
    "FloatField",
    "DateField",
    "TimeField",
    "NaiveTimeField",
    "DateTimeField",
    "NaiveDateTimeField",
    "HttpDateTimeField",
    "TimeStampField",
    "EmailField",
    "IPv4Field",
    "IPv6Field",
    "IPv46Field",
    "UUIDField",
    "DictField",
    "ObjectField",
    "ArrayField",
    "TypedArrayField",
    "TypedListField",
    "TypedDictField",
    "TypedObjectField",
    "NotProvided",
)


if six.PY3:
    long = int


class NotProvided:
    pass


# Backwards compatibility
NOT_PROVIDED = NotProvided


class Field(BaseField):
    """
    Base class for fields.
    """

    default_validators = []
    default_error_messages = {
        "invalid_choice": "Value %r is not a valid choice.",
        "null": "This field cannot be null.",
        "required": "This field is required.",
    }
    data_type_name = None

    def __init__(
        self,
        verbose_name=None,
        verbose_name_plural=None,
        name=None,
        null=False,
        choices=None,
        use_default_if_not_provided=False,
        default=NotProvided,
        help_text="",
        validators=None,
        error_messages=None,
        is_attribute=False,
        doc_text="",
        key=False,
    ):
        """
        Initialisation of a Field.

        :param verbose_name: Display name of field.
        :param verbose_name_plural: Plural display name of field.
        :param name: Name of the serialised field.
        :param null: This value can be null/None.
        :param choices: Collection of valid choices for this field.
        :param use_default_if_not_provided: If a value is not provided during mapping, use the default value.
        :param default: Default value for this field.
        :param help_text: Deprecated; Help text to describe this field.
        :param validators: Additional validators, these should be a callable that takes a single value.
        :param error_messages: Dictionary that overrides error messages (or provides additional messages for custom
            validation).
        :param is_attribute: Special flag for codecs that support attributes on nodes (ie XML)
        :param doc_text: Documentation for the field, replaces help text
        :param key: Mark this field as a key field (used to generated a unique identifier).

        """
        super(Field, self).__init__(
            verbose_name, verbose_name_plural, name, doc_text or help_text
        )

        self.null, self.choices = null, choices
        self.default, self.use_default_if_not_provided = (
            default,
            use_default_if_not_provided,
        )
        self.validators = self.default_validators + (validators or [])
        self.is_attribute = is_attribute
        self.key = key

        messages = {}
        for c in reversed(self.__class__.__mro__):
            messages.update(getattr(c, "default_error_messages", {}))
        messages.update(error_messages or {})
        self.error_messages = messages

        self.resource = None

    def __deepcopy__(self, memodict):
        # We don't have to deepcopy very much here, since most things are not
        # intended to be altered after initial creation.
        obj = copy.copy(self)
        memodict[id(self)] = obj
        return obj

    @lazy_property
    def choice_values(self):
        """
        Choice values to allow choices to simplify checking if a choice is valid.
        """
        if self.choices is not None:
            return tuple(c[0] for c in self.choices)

    @property
    def choices_doc_text(self):
        # type: () -> Sequence[Tuple[str, str]]
        """
        Choices converted for documentation purposes.
        """
        return self.choices

    def contribute_to_class(self, cls, name):
        self.set_attributes_from_name(name)
        self.resource = cls
        getmeta(cls).add_field(self)

    def to_python(self, value):
        """
        Converts the input value into the expected Python data type, raising
        odin.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.
        """
        raise NotImplementedError()

    def run_validators(self, value):
        if value in EMPTY_VALUES:
            return

        errors = []
        for v in self.validators:
            try:
                v(value)
            except registration.get_validation_error_list() as e:
                handler = registration.get_validation_error_handler(e)
                handler(e, self, errors)
        if errors:
            raise exceptions.ValidationError(errors)

    def validate(self, value):
        if (
            self.choice_values
            and (value not in EMPTY_VALUES)
            and (value not in self.choice_values)
        ):
            msg = self.error_messages["invalid_choice"] % value
            raise exceptions.ValidationError(msg)

        if not self.null and value is None:
            raise exceptions.ValidationError(self.error_messages["null"])

    def clean(self, value):
        """
        Convert the value's type and run validation. Validation errors
        from to_python and validate are propagated. The correct value is
        returned if no error is raised.
        """
        if value is NotProvided:
            value = self.get_default() if self.use_default_if_not_provided else None
        value = self.to_python(value)
        self.validate(value)
        self.run_validators(value)
        return value

    def has_default(self):
        """
        Returns a boolean of whether this field has a default value.
        """
        return self.default is not NotProvided

    def get_default(self):
        """
        Returns the default value for this field.
        """
        if self.has_default():
            if callable(self.default):
                return self.default()
            return self.default
        return None

    def value_to_object(self, obj, data):
        """
        Assign a value to an object
        """
        setattr(obj, self.attname, data)


class BooleanField(Field):
    default_error_messages = {"invalid": "'%s' value must be either True or False."}
    true_strings = ("t", "true", "y", "yes", "on", "1", "✓")
    false_strings = ("f", "false", "n", "no", "off", "0")
    data_type_name = "Boolean"

    def to_python(self, value):
        if value is None:
            return None
        if value in (True, False):
            # if value is 1 or 0 then it's equal to True or False, but we want
            # to return a true bool for semantic reasons.
            return bool(value)
        if isinstance(value, six.string_types):
            lvalue = value.lower()
            if lvalue in self.true_strings:
                return True
            if lvalue in self.false_strings:
                return False
        msg = self.error_messages["invalid"] % str(value)
        raise exceptions.ValidationError(msg)


class StringField(Field):
    """
    A string.

    StringField has two extra arguments:

    :py:attr:`StringField.empty`
        The string can be empty. This is similar to None but captures an empty string in addition to ``None``.
        The default state of this flag is ``False``.

    :py:attr:`StringField.max_length`
        The maximum length (in characters) of the field. The ``max_length`` value is enforced Odin’s validation.
    """

    data_type_name = "String"

    def __init__(self, max_length=None, empty=None, **options):
        super(StringField, self).__init__(**options)
        self.max_length = max_length

        # Mirror null is not explicitly defined
        if empty is None:
            empty = options.get("null", False)
        self.empty = empty

        if max_length is not None:
            self.validators.append(MaxLengthValidator(max_length))

    def to_python(self, value):
        if value is None or isinstance(value, six.string_types):
            return value
        return str(value)

    def validate(self, value):
        if not self.empty and value == "":
            raise exceptions.ValidationError(self.error_messages["null"])

        super(StringField, self).validate(value)


class UrlField(StringField):
    data_type_name = "URL"

    def __init__(self, **options):
        options.setdefault("validators", []).append(validate_url)
        super(UrlField, self).__init__(**options)


class ScalarField(Field):
    scalar_type = int

    def __init__(self, min_value=None, max_value=None, **options):
        super(ScalarField, self).__init__(**options)
        self.min_value = min_value
        if min_value is not None:
            self.validators.append(MinValueValidator(min_value))
        self.max_value = max_value
        if max_value is not None:
            self.validators.append(MaxValueValidator(max_value))

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        try:
            return self.scalar_type(value)
        except (TypeError, ValueError):
            msg = self.error_messages["invalid"] % value
            raise exceptions.ValidationError(msg)


class IntegerField(ScalarField):
    default_error_messages = {
        "invalid": "'%s' value must be a integer.",
    }
    data_type_name = "Integer"


class FloatField(ScalarField):
    default_error_messages = {
        "invalid": "'%s' value must be a float.",
    }
    data_type_name = "Float"
    scalar_type = float


class _IsoFormatMixin(BaseField):
    def as_string(self, value):
        """
        Generate a string representation of a field.

        :param value:
        :rtype: str

        """
        if value:
            return value.isoformat()


class DateField(_IsoFormatMixin, Field):
    """
    Field that handles date values encoded as a string.

    The format of the string is that defined by ISO-8601.

    """

    default_error_messages = {
        "invalid": "Not a valid date string.",
    }
    data_type_name = "ISO-8601 Date"

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.datetime):
            return value.date()
        if isinstance(value, datetime.date):
            return value
        try:
            return datetimeutil.parse_iso_date_string(value)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)


class TimeField(_IsoFormatMixin, Field):
    """
    Field that handles time values encoded as a string.

    The format of the string is that defined by ISO-8601.

    Use the ``assume_local`` flag to customise how naive (datetime values with
    no timezone) are handled and also how dates are decoded. If
    ``assume_local`` is True naive dates are assumed to represent the current
    system timezone.

    """

    default_error_messages = {
        "invalid": "Not a valid time string.",
    }
    data_type_name = "ISO-8601 Time"

    def __init__(self, assume_local=False, **options):
        super(TimeField, self).__init__(**options)
        self.assume_local = assume_local

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.time):
            return value
        default_timezone = datetimeutil.local if self.assume_local else datetimeutil.utc
        try:
            return datetimeutil.parse_iso_time_string(value, default_timezone)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)


class NaiveTimeField(_IsoFormatMixin, Field):
    """
    Field that handles time values encoded as a string.

    The format of the string is that defined by ISO-8601.

    The naive time field differs from :py:`~TimeField` in the handling of the
    timezone, a timezone will not be applied if one is not specified.

    Use the ``ignore_timezone`` flag to have any timezone information ignored
    when decoding the time field.

    """

    default_error_messages = {
        "invalid": "Not a valid time string.",
    }
    data_type_name = "Naive ISO-8601 Time"

    def __init__(self, ignore_timezone=False, **options):
        super(NaiveTimeField, self).__init__(**options)
        self.ignore_timezone = ignore_timezone

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.time):
            if value.tzinfo and self.ignore_timezone:
                return value.replace(tzinfo=None)
            return value
        default_timezone = datetimeutil.IgnoreTimezone if self.ignore_timezone else None
        try:
            return datetimeutil.parse_iso_time_string(value, default_timezone)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)

    def prepare(self, value):
        """
        Prepare for serialisation

        :param value:
        :return:

        """
        if (value is not None) and self.ignore_timezone and (value.tzinfo is not None):
            # Strip the timezone
            value = value.replace(tzinfo=None)
        return value


class DateTimeField(_IsoFormatMixin, Field):
    """
    Field that handles datetime values encoded as a string.

    The format of the string is that defined by ISO-8601.

    Use the ``assume_local`` flag to customise how naive (datetime values
    with no timezone) are handled and also how dates are decoded. If
    ``assume_local`` is True naive dates are assumed to represent the current
    system timezone.

    """

    default_error_messages = {
        "invalid": "Not a valid datetime string.",
    }
    data_type_name = "ISO-8601 DateTime"

    def __init__(self, assume_local=False, **options):
        super(DateTimeField, self).__init__(**options)
        self.assume_local = assume_local

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.datetime):
            return value
        default_timezone = datetimeutil.local if self.assume_local else datetimeutil.utc
        try:
            return datetimeutil.parse_iso_datetime_string(value, default_timezone)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)


class NaiveDateTimeField(_IsoFormatMixin, Field):
    """
    Field that handles datetime values encoded as a string.

    The format of the string is that defined by ISO-8601.

    The naive time field differs from :py:`~DateTimeField` in the handling of the
    timezone, a timezone will not be applied if one is not specified.

    Use the ``ignore_timezone`` flag to have any timezone information ignored
    when decoding the time field.

    """

    default_error_messages = {
        "invalid": "Not a valid datetime string.",
    }
    data_type_name = "Naive ISO-8601 DateTime"

    def __init__(self, ignore_timezone=False, **options):
        super(NaiveDateTimeField, self).__init__(**options)
        self.ignore_timezone = ignore_timezone

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.datetime):
            if value.tzinfo and self.ignore_timezone:
                return value.replace(tzinfo=None)
            return value
        default_timezone = datetimeutil.IgnoreTimezone if self.ignore_timezone else None
        try:
            return datetimeutil.parse_iso_datetime_string(value, default_timezone)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)

    def prepare(self, value):
        """
        Prepare for serialisation

        :param value:
        :return:

        """
        if (value is not None) and self.ignore_timezone and (value.tzinfo is not None):
            # Strip the timezone
            value = value.replace(tzinfo=None)
        return value


class HttpDateTimeField(Field):
    """
    Field that handles datetime values encoded as a string.

    The format of the string is that defined by ISO-1123.

    """

    default_error_messages = {
        "invalid": "Not a valid HTTP datetime string.",
    }
    data_type_name = "ISO-1123 DateTime"

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.datetime):
            return value
        try:
            return datetimeutil.parse_http_datetime_string(value)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)

    def as_string(self, value):
        """
        Generate a string representation of a field.

        :param value:
        :rtype: str

        """
        if value is not None:
            return datetimeutil.to_http_datetime_string(value)


class TimeStampField(Field):
    """
    Field that handles datetime values encoding as the number of seconds since the UNIX epoch.

    A UNIX timestamp should always be calculated relative to UTC.

    """

    default_error_messages = {
        "invalid": "Not a valid UNIX timestamp.",
    }
    data_type_name = "Integer"

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, datetime.datetime):
            return value
        try:
            return datetime.datetime.fromtimestamp(long(value), tz=datetimeutil.utc)
        except ValueError:
            pass
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)

    def prepare(self, value):
        if value in EMPTY_VALUES:
            return
        if isinstance(value, six.integer_types):
            return long(value)
        if isinstance(value, datetime.datetime):
            return datetimeutil.to_timestamp(value)


class DictField(Field):
    default_error_messages = {
        "invalid": "Must be a dict.",
    }
    data_type_name = "Dict"

    def __init__(self, **options):
        options.setdefault("default", dict)
        super(DictField, self).__init__(**options)

    def to_python(self, value):
        if value is None:
            return value
        try:
            val = dict(value)
            return val
        except (TypeError, ValueError):
            msg = self.error_messages["invalid"]
            raise exceptions.ValidationError(msg)


ObjectField = DictField


class ListField(Field):
    default_error_messages = {
        "invalid": "Must be an array.",
    }
    data_type_name = "List"

    def __init__(self, **options):
        options.setdefault("default", list)
        super(ListField, self).__init__(**options)

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, (list, tuple)):
            return value
        msg = self.error_messages["invalid"]
        raise exceptions.ValidationError(msg)


ArrayField = ListField


class TypedListField(ListField):
    @staticmethod
    def data_type_name(instance):
        type_name = instance.field.data_type_name
        if callable(type_name):
            type_name = type_name(instance.field)
        return "List<{0}>".format(type_name)

    def __init__(self, field, **options):
        self.field = field
        super(TypedListField, self).__init__(**options)

    def to_python(self, value):
        value = super(TypedListField, self).to_python(value)
        if not value:
            return value

        value_list = []
        errors = {}
        for idx, item in enumerate(value):
            try:
                value_list.append(self.field.to_python(item))
            except exceptions.ValidationError as ve:
                errors[idx] = ve.error_messages

        if errors:
            raise exceptions.ValidationError(errors)

        return value_list

    def prepare(self, value):
        if isinstance(value, (tuple, list)):
            prepare = self.field.prepare
            return [prepare(i) for i in value]
        return value


TypedArrayField = TypedListField


class TypedDictField(DictField):
    """
    Dict field with both key and value fixed to a specific types. By default the key field is assumed to be a string.

    Usage::

        # Dict with key and value fields as string.
        TypedDictField(key_field=StringField(), value_field=StringField())

    """

    @staticmethod
    def data_type_name(instance):
        key_type_name = instance.key_field.data_type_name
        if callable(key_type_name):
            key_type_name = key_type_name(instance.key_field)

        value_type_name = instance.value_field.data_type_name
        if callable(value_type_name):
            value_type_name = value_type_name(instance.value_field)

        return "Dict<{0}, {1}>".format(key_type_name, value_type_name)

    def __init__(self, value_field, key_field=StringField(), **options):
        self.key_field = key_field
        self.value_field = value_field
        super(TypedDictField, self).__init__(**options)

    def to_python(self, value):
        value = super(TypedDictField, self).to_python(value)
        if not value:
            return value

        value_dict = {}
        key_errors = []
        value_errors = {}
        for key, value in value.items():
            try:
                key = self.key_field.to_python(key)
            except exceptions.ValidationError as ve:
                key_errors += ve.error_messages

            # If we have key errors no point checking values any more.
            if key_errors:
                continue

            try:
                value_dict[key] = self.value_field.to_python(value)
            except exceptions.ValidationError as ve:
                value_errors[key] = ve.error_messages

        if key_errors:
            raise exceptions.ValidationError(key_errors)
        if value_errors:
            raise exceptions.ValidationError(value_errors)

        return value_dict

    def validate(self, value):
        super(TypedDictField, self).validate(value)

        if value in EMPTY_VALUES:
            return

        key_errors = []
        value_errors = {}
        for key, value in value.items():
            try:
                self.key_field.validate(key)
            except exceptions.ValidationError as ve:
                key_errors += ve.error_messages

            # If we have key errors no point checking values any more.
            if key_errors:
                continue

            try:
                self.value_field.validate(value)
            except exceptions.ValidationError as ve:
                value_errors[key] = ve.error_messages

        if key_errors:
            raise exceptions.ValidationError(key_errors)
        if value_errors:
            raise exceptions.ValidationError(value_errors)

    def run_validators(self, value):
        super(TypedDictField, self).run_validators(value)

        if value in EMPTY_VALUES:
            return

        key_errors = []
        value_errors = {}
        for key, value in value.items():
            try:
                key = self.key_field.run_validators(key)
            except exceptions.ValidationError as ve:
                key_errors += ve.error_messages

            # If we have key errors no point checking values any more.
            if key_errors:
                continue

            try:
                self.value_field.run_validators(value)
            except exceptions.ValidationError as ve:
                value_errors[key] = ve.error_messages

        if key_errors:
            raise exceptions.ValidationError(key_errors)
        if value_errors:
            raise exceptions.ValidationError(value_errors)


TypedObjectField = TypedDictField


class EmailField(StringField):
    """
    An Email address.

    Validates that a string represents a valid Email address.

    """

    data_type_name = "Email"

    def __init__(self, **options):
        options.setdefault("validators", []).append(validate_email_address)
        super(EmailField, self).__init__(**options)


class IPv4Field(StringField):
    """
    An IPv4 address.

    Validates that a string represents a valid IPv4 address format.

    """

    data_type_name = "IPv4"

    def __init__(self, **options):
        options.setdefault("validators", []).append(validate_ipv4_address)
        super(IPv4Field, self).__init__(**options)


class IPv6Field(StringField):
    """
    An IPv6 address.

    Validates that a string represents a valid IPv6 address format.

    """

    data_type_name = "IPv6"

    def __init__(self, **options):
        options.setdefault("validators", []).append(validate_ipv6_address)
        super(IPv6Field, self).__init__(**options)


class IPv46Field(StringField):
    """
    An IPv4 or IPv6 address.

    Validates that a string represents a valid IPv4 or IPv6 address format.

    """

    data_type_name = "IPv46"

    def __init__(self, **options):
        options.setdefault("validators", []).append(validate_ipv46_address)
        super(IPv46Field, self).__init__(**options)


class UUIDField(Field):
    """
    An universally unique identifier.

    Validates that the string represents a universally unique identifier.
    """

    data_type_name = "UUID"

    def __init__(self, **options):
        super(UUIDField, self).__init__(**options)

    def to_python(self, value):
        if value is None:
            return

        if isinstance(value, uuid.UUID):
            return value

        elif isinstance(value, six.binary_type):
            if len(value) == 16:
                return uuid.UUID(bytes=value)

            try:
                value = value.decode("utf-8")
            except UnicodeDecodeError as e:
                raise exceptions.ValidationError(e.args[0], code="invalid")

        elif isinstance(value, six.integer_types):
            try:
                return uuid.UUID(int=value)
            except ValueError as e:
                raise exceptions.ValidationError(e.args[0], code="invalid")

        elif isinstance(value, (tuple, list)):
            try:
                return uuid.UUID(fields=value)
            except ValueError as e:
                raise exceptions.ValidationError(e.args[0], code="invalid")

        elif not isinstance(value, six.text_type):
            value = six.text_type(value)

        try:
            return uuid.UUID(value)
        except ValueError as e:
            raise exceptions.ValidationError(e.args[0], code="invalid")



# Disable logging if an explicit handler is not added
import logging

logging.getLogger("odin.registration").addHandler(logging.NullHandler())

from odin.fields import *  # noqa
from odin.fields.composite import *  # noqa
from odin.fields.future import *  # noqa
from odin.fields.virtual import *  # noqa
from odin.mapping import *  # noqa
from odin.resources import Resource  # noqa
from odin.adapters import ResourceAdapter  # noqa
from odin.proxy import ResourceProxy  # noqa

__authors__ = "Tim Savage <tim@savage.company>"
__copyright__ = "Copyright (C) 2021 Tim Savage"



import pytest

from odin.resources import create_resource_from_dict
from odin.utils import getmeta
from odin import proxy

from .resources import Book, BookProxy


class MockObject(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TestFieldProxyDescriptor(object):
    @pytest.fixture
    def proxy_obj(self):
        class Target(object):
            def __init__(self, shadow):
                self._shadow = shadow

        proxy.FieldProxyDescriptor().contribute_to_class(Target, 'foo')
        proxy.FieldProxyDescriptor().contribute_to_class(Target, 'bar')
        proxy.FieldProxyDescriptor(readonly=True).contribute_to_class(Target, 'eek')

        return Target

    def test_get(self, proxy_obj):
        target = proxy_obj(MockObject(foo='123', eek='789'))

        assert target.foo == '123'
        with pytest.raises(AttributeError):
            print(target.bar)

    def test_set(self, proxy_obj):
        shadow = MockObject(foo='123', eek='789')
        target = proxy_obj(shadow)

        target.foo = '321'
        target.bar = '654'

        assert shadow.foo == '321'
        assert shadow.bar == '654'

    def test_set__readonly(self, proxy_obj):
        shadow = MockObject(foo='123', eek='789')
        target = proxy_obj(shadow)

        with pytest.raises(AttributeError):
            target.eek = 321

        assert '789' == shadow.eek


class TestResourceProxyMeta(object):
    def test_no_meta(self):
        with pytest.raises(AttributeError):
            class SampleProxy(proxy.ResourceProxy):
                pass

    def test_no_resource(self):
        with pytest.raises(AttributeError):
            class SampleProxy(proxy.ResourceProxy):
                class Meta:
                    pass

    def test_repr(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                resource = Book

        assert '<Proxy of <Options for library.Book>>' == repr(getmeta(SampleProxy))

    def test_field_sorting(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                field_sorting = True
                resource = Book

        assert [f.attname for f in getmeta(SampleProxy).fields] == [
            'title', 'isbn', 'num_pages', 'rrp', 'fiction', 'genre', 'published', 'authors', 'publisher']

    def test_field_sorting__custom(self):
        class SampleProxy(proxy.ResourceProxy):
            class Meta:
                def field_sorting(fields):
                    return sorted(fields, key=lambda f: f.attname)
                resource = Book

        assert [f.attname for f in getmeta(SampleProxy).fields] == [
            'authors', 'fiction', 'genre', 'isbn', 'num_pages', 'published', 'publisher', 'rrp', 'title']


class TestResourceProxyOptions(object):
    def test_options_type(self):
        actual = getmeta(BookProxy)
        assert isinstance(actual, proxy.ResourceProxyOptions)

    def test_invalid_option(self):
        with pytest.raises(TypeError):
            class SampleProxy(proxy.ResourceProxy):
                class Meta:
                    resource = Book
                    eek = 'boo!'

    @pytest.mark.parametrize('attr, actual', (
        ('key_field_names', ('isbn',)),
        ('verbose_name', 'Book Summary'),
        ('verbose_name_plural', 'Book Summarys'),
        ('include', ('title', 'isbn', 'num_pages', 'rrp')),
        ('name_space', 'the.other.library'),
    ))
    def test_options(self, attr, actual):
        target = getmeta(BookProxy)
        assert actual == getattr(target, attr)


class TestResourceProxy(object):
    def test_shadow(self):
        target = BookProxy()
        actual = target.get_shadow()

        assert isinstance(actual, Book)

    def test_proxy(self):
        actual = Book(title='1984')
        target = BookProxy.proxy(actual)

        assert target.title == '1984'
        assert actual is target.get_shadow()

    def test_proxy__iterable(self):
        actual = [Book(title='1984'), Book(title='1985 - The man strikes back'), Book(title='1986 - Return of the man')]
        target = BookProxy.proxy(actual)

        assert isinstance(target, proxy.ResourceProxyIter)
        assert ['1984', '1985 - The man strikes back', '1986 - Return of the man'] == [t.title for t in target]

    def test_proxy__unknown(self):
        with pytest.raises(TypeError):
            BookProxy.proxy(123)

    def test_extra_property(self):
        book = Book(rrp=256)
        target = BookProxy.proxy(book)

        assert target.expensive

    def test_create_resource_from_dict(self):
        actual = create_resource_from_dict({'title': '1984', 'isbn': '1234567', 'num_pages': 1}, BookProxy)

        assert actual.title == '1984'



# -*- coding: utf-8 -*-
import arrow
import datetime

from odin import exceptions, datetimeutil
from odin.fields import Field
from odin.validators import EMPTY_VALUES

__all__ = ('ArrowField',)


class ArrowField(Field):
    """
    Field that handles datetime values encoded as a string utilising the Arrow
    date time interface.

    The format of the string is that defined by ISO-8601.

    Use the ``assume_local`` flag to customise how naive (datetime values
    with no timezone) are handled and also how dates are decoded. If
    ``assume_local`` is True naive dates are assumed to represent the current
    system timezone.

    """
    default_error_messages = {
        'invalid': "Not a valid datetime string.",
    }
    data_type_name = "ISO-8601 DateTime"

    def __init__(self, assume_local=False, **options):
        super(ArrowField, self).__init__(**options)
        self.assume_local = assume_local

    def to_python(self, value):
        if value in EMPTY_VALUES:
            return

        if isinstance(value, arrow.Arrow):
            return value

        default_timezone = datetimeutil.local if self.assume_local else datetimeutil.utc

        if isinstance(value, datetime.datetime):
            if value.tzinfo:
                return arrow.Arrow.fromdatetime(value)
            else:
                return arrow.Arrow.fromdatetime(value, default_timezone)

        try:
            return arrow.Arrow.fromdatetime(datetimeutil.parse_iso_datetime_string(value, default_timezone))
        except ValueError:
            pass

        try:
            return arrow.Arrow.fromdate(datetimeutil.parse_iso_date_string(value))
        except ValueError:
            pass
        msg = self.error_messages['invalid']
        raise exceptions.ValidationError(msg)



from __future__ import absolute_import

from enum import Enum
from typing import TypeVar, Optional, Any, Type  # noqa

from odin.exceptions import ValidationError
from . import Field

__all__ = ("EnumField",)


ET = TypeVar("ET", Enum, Enum)


class EnumField(Field):
    """
    Field for handling Python enums.
    """

    data_type_name = "Enum"

    def __init__(self, enum, **options):
        # type: (Type[ET], **Any) -> None

        # Generate choices structure from choices
        choices = options.pop("choices", None)
        options["choices"] = tuple((e, e.name) for e in choices or enum)

        super(EnumField, self).__init__(**options)
        self.enum = enum

    @property
    def choices_doc_text(self):
        """
        Choices converted for documentation purposes.
        """
        return tuple((v.value, n) for v, n in self.choices)

    def to_python(self, value):
        # type: (Any) -> Optional[ET]
        if value is None:
            return

        # Attempt to convert
        try:
            return self.enum(value)
        except ValueError:
            # If value is an empty string return None
            # Do this check here to support enums that define an option using
            # an empty string.
            if value == "":
                return
            raise ValidationError(self.error_messages["invalid_choice"] % value)

    def prepare(self, value):
        # type: (Optional[ET]) -> Any
        if (value is not None) and isinstance(value, self.enum):
            return value.value



# -*- coding: utf-8 -*-
import pytest
import odin
from . import resources


class TestResourceBase(object):
    def test_virtual_fields(self):
        target = resources.FromResource()

        assert ['constant_field'] == [f.name for f in target._meta.virtual_fields]

    def test_virtual_field_inheritance(self):
        target = resources.InheritedResource()

        assert ['calculated_field', 'constant_field'] == [f.name for f in target._meta.virtual_fields]

    def test_fields(self):
        target = resources.FromResource()

        assert (['title', 'count', 'child', 'children', 'excluded1', 'from_field1', 'from_field2',
                 'from_field3', 'from_field4', 'same_but_different', 'from_field_c1', 'from_field_c2',
                 'from_field_c3', 'from_field_c4', 'not_auto_c5', 'comma_separated_string'] ==
                [f.name for f in target._meta.fields])

    def test_field_inheritance(self):
        target = resources.InheritedResource()

        assert (['name', 'title', 'count', 'child', 'children', 'excluded1', 'from_field1', 'from_field2',
                 'from_field3', 'from_field4', 'same_but_different', 'from_field_c1', 'from_field_c2',
                 'from_field_c3', 'from_field_c4', 'not_auto_c5', 'comma_separated_string'] ==
                [f.name for f in target._meta.fields])

    def test_field_multi_inheritance(self):
        target = resources.MultiInheritedResource()

        assert (['name', 'title', 'count', 'child', 'children', 'excluded1', 'from_field1', 'from_field2',
                 'from_field3', 'from_field4', 'same_but_different', 'from_field_c1', 'from_field_c2',
                 'from_field_c3', 'from_field_c4', 'not_auto_c5', 'comma_separated_string'] ==
                [f.name for f in target._meta.fields])

    def test_key_field_does_not_exist(self):
        """Ensure an exception is raised if specified key_field is not a member of the resource"""
        with pytest.raises(AttributeError):
            class BadResource(odin.Resource):
                class Meta:
                    key_field_name = 'missing_field'

                field_1 = odin.StringField()

    def test_key_fields_from_field_attributes(self):
        """
        Ensure that multiple keys can be defined from fields and in the right order.
        """
        assert ['name', 'genre'] == [f.attname for f in resources.OldBook._meta.key_fields]



# -*- coding: utf-8 -*-
import datetime
from odin import datetimeutil
from odin.compatibility import deprecated

STRING_TYPES = {}


def register_string_type(data_type, serialise_method):
    """
    Register a data type serialise method that results in a string.

    :param data_type: a type instance that string data type should match.
    :param serialise_method: a method that accepts a single value that will return a string representation of
        the serialised type.

    """
    STRING_TYPES[data_type] = serialise_method


def date_iso_format(value):
    # type: (datetime.date) -> str
    """
    Serialise a datetime.date to ISO string format.
    """
    return value.isoformat()


@deprecated("Defaulting the timezone should be preformed by the fields only, this confuses things. "
            "Most codecs have already migrated to not use this class.")
class DatetimeIsoFormat(object):
    """
    Serialise a datetime.time or datetime.datetime to ISO string format.
    """
    def __init__(self, default_timezone=datetimeutil.local):
        self.default_timezone = default_timezone

    def __call__(self, value):
        # type: (datetime.time | datetime.datetime) -> str
        if value.tzinfo is None:
            value = value.replace(tzinfo=self.default_timezone)
        return value.isoformat()

TimeIsoFormat = DatetimeIsoFormat


def datetime_iso_format(value):
    """
    Serialise a datetime.datetime to ISO string format.
    """
    # type: (datetime.datetime) -> str
    return value.isoformat()


def time_iso_format(value):
    """
    Serialise a datetime.time to ISO string format.
    """
    # type: (datetime.time) -> str
    return value.isoformat()


@deprecated("Defaulting the timezone should be preformed by the fields only, this confuses things. "
            "Most codecs have already migrated to not use this class.")
class DatetimeEcmaFormat(object):
    """
    Serialize a datetime object into the ECMA defined format.
    """
    input_type = datetime.datetime

    def __init__(self, assume_local=True, default_timezone=datetimeutil.local):
        self.default_timezone = datetimeutil.local if assume_local else default_timezone

    def __call__(self, value):
        return datetimeutil.to_ecma_datetime_string(value, self.default_timezone)

datetime_ecma_format = DatetimeEcmaFormat()



import arrow
import datetime
import pytest

from odin.datetimeutil import utc
from odin.exceptions import ValidationError

try:
    from odin.contrib.arrow.fields import ArrowField
except ImportError:
    pass  # Arrow is not available
else:

    class TestArrowField(object):
        @pytest.mark.parametrize('kwargs, value, expected', (
            ({'null': True}, None, None),
            ({}, '2013-11-24', arrow.Arrow(2013, 11, 24)),
            ({}, '2013-11-24T18:43:00.000', arrow.Arrow(2013, 11, 24, 18, 43)),
            ({}, '2013-11-24T18:43:00.000Z', arrow.Arrow(2013, 11, 24, 18, 43)),
            ({}, arrow.Arrow(2013, 11, 24, 18, 43), arrow.Arrow(2013, 11, 24, 18, 43)),
            ({}, datetime.datetime(2013, 11, 24, 18, 43), arrow.Arrow(2013, 11, 24, 18, 43)),
            ({}, datetime.datetime(2013, 11, 24, 18, 43, 0, 0, utc), arrow.Arrow(2013, 11, 24, 18, 43)),
        ))
        def test_arrowfield__clean__valid(self, kwargs, value, expected):
            target = ArrowField(**kwargs)
            actual = target.clean(value)

            assert actual == expected

        @pytest.mark.parametrize('kwargs, value', (
            ({}, None),
            ({}, 123),
            ({}, '2013-11-'),
        ))
        def test_arrowfield__clean__invalid(self, kwargs, value):
            with pytest.raises(ValidationError):
                ArrowField(**kwargs).clean(value)



"""
Resource Proxy
~~~~~~~~~~~~~~

The resource proxy is an object that provides an alternate interface to a shadowed Resource.

This could be as a means of providing a summary object, or for adding additional properties.

"""

import six

# Typing includes
from typing import List, Union  # noqa

from odin import registration

from odin.bases import TypedResourceIterable
from odin.resources import ResourceOptions, ResourceBase
from odin.utils import getmeta, filter_fields, lazy_property

EMPTY = tuple()


class FieldProxyDescriptor(object):
    """
    Descriptor to proxy field to underlying resource.
    """
    __slots__ = ('attname', 'readonly')

    def __init__(self, readonly=False):
        self.attname = None  # type: str
        self.readonly = readonly

    def __get__(self, instance, owner):
        shadow = getattr(instance, '_shadow')
        return getattr(shadow, self.attname)

    def __set__(self, instance, value):
        if self.readonly:
            raise AttributeError("can't set attribute")
        shadow = getattr(instance, '_shadow')
        return setattr(shadow, self.attname, value)

    def contribute_to_class(self, cls, name):
        self.attname = name
        setattr(cls, name, self)


class ResourceProxyOptions(ResourceOptions):
    META_OPTION_NAMES = (
        'include', 'exclude', 'readonly',
        # Overrides
        'name', 'name_space', 'namespace', 'verbose_name', 'verbose_name_plural', 'doc_group',
        'type_field', 'key_field_name', 'key_field_names', 'field_sorting',
    )

    def __init__(self, meta):
        super(ResourceProxyOptions, self).__init__(meta)
        self.resource = None
        self.shadow = None
        self.include = []
        self.exclude = []
        self.readonly = []

    def __repr__(self):
        return '<Proxy of {!r}>'.format(getmeta(self.resource))

    def contribute_to_class(self, cls, _):
        cls._meta = self
        cls_name = cls.__name__
        self.name = cls_name
        self.class_name = "{}.{}".format(cls.__module__, cls_name)

        # Get and filter meta attributes
        meta_attrs = self.meta.__dict__.copy()
        for name in self.meta.__dict__:
            if name.startswith('_'):
                del meta_attrs[name]

        # Get the required resource object
        self.resource = meta_attrs.pop('resource', None)
        if not self.resource:
            raise AttributeError('`resource` has not been defined.')
        self.shadow = shadow = getmeta(self.resource)

        # Extract all meta options and fetch from shadow if not defined
        proxy_attrs = {
            'name': cls_name,
            'verbose_name': cls_name.replace('_', ' ').strip('_ '),
        }
        for attr_name in self.META_OPTION_NAMES:
            if attr_name in meta_attrs:
                value = meta_attrs.pop(attr_name)

                if attr_name == 'verbose_name':
                    # If defined generate pluralised form base on this name.
                    if 'verbose_name_plural' not in proxy_attrs:
                        proxy_attrs['verbose_name_plural'] = value + 's'

                elif attr_name == 'namespace':
                    # Allow meta to be defined as namespace
                    attr_name = 'name_space'

                elif attr_name == 'key_field_name':
                    # Remap to key_field names
                    attr_name = 'key_field_names'
                    value = [value]

                proxy_attrs[attr_name] = value

            elif hasattr(shadow, attr_name):
                proxy_attrs.setdefault(attr_name, getattr(shadow, attr_name))

        # Any leftover attributes must be invalid.
        if meta_attrs != {}:
            raise TypeError("'class Meta' got invalid attribute(s): %s" % ','.join(meta_attrs.keys()))
        del self.meta

        # Apply to self
        for attr_name, value in proxy_attrs.items():
            setattr(self, attr_name, value)

    @lazy_property
    def readonly_fields(self):
        """
        Fields that can only be read from.
        """
        return tuple(f for f in self.fields if f.attname in self.readonly)

    @lazy_property
    def init_fields(self):
        """
        Fields used in the resource init
        """
        return tuple(f for f in self.fields if f.attname not in self.readonly)


class ResourceProxyType(type):
    meta_options = ResourceProxyOptions

    def __new__(mcs, name, bases, attrs):
        super_new = super(ResourceProxyType, mcs).__new__

        # attrs will never be empty for classes declared in the standard way
        # (ie. with the `class` keyword). This is quite robust.
        if name == 'NewBase' and attrs == {}:
            return super_new(mcs, name, bases, attrs)

        parents = [
            b for b in bases if
            isinstance(b, ResourceProxyType) and not (b.__name__ == 'NewBase' and b.__mro__ == (b, object))
        ]
        if not parents:
            # If this isn't a subclass of Resource, don't do anything special.
            return super_new(mcs, name, bases, attrs)

        # Create the class.
        module = attrs.pop('__module__')
        new_class = super_new(mcs, name, bases, {'__module__': module})
        attr_meta = attrs.pop('Meta', None)
        if not attr_meta:
            meta = getattr(new_class, 'Meta', None)
        else:
            meta = attr_meta

        new_meta = mcs.meta_options(meta)
        new_class.add_to_class('_meta', new_meta)

        # Add all attributes to the class.
        for obj_name, obj in attrs.items():
            new_class.add_to_class(obj_name, obj)

        # Determine which fields will be shadowed.
        field_names, readonly = filter_fields(
            new_meta.shadow.field_map,
            new_meta.include,
            new_meta.exclude,
            new_meta.readonly
        )

        # Map field names
        new_meta.fields = [f for f in new_meta.shadow.fields if f.attname in field_names]

        # Sort the fields
        if new_meta.field_sorting:
            if callable(new_meta.field_sorting):
                new_meta.fields = new_meta.field_sorting(new_meta.fields)
            else:
                new_meta.fields = sorted(new_meta.fields, key=hash)

        # Generate field descriptors
        for field in new_meta.fields:
            new_class.add_to_class(field.attname, FieldProxyDescriptor(field in new_meta.readonly_fields))

        # If a key_field is defined ensure it exists
        if new_meta.key_field_names:
            for field_name in new_meta.key_field_names:
                if field_name not in new_meta.field_map:
                    raise AttributeError('Key field `{0}` is not exist on this resource.'.format(field_name))

        # Register resource
        registration.register_resources(new_class)

        # Because of the way imports happen (recursively), we may or may not be
        # the first time this model tries to register with the framework. There
        # should only be one class for each model, so we always return the
        # registered version.
        return registration.get_resource(new_meta.resource_name)

    def add_to_class(cls, name, value):
        if hasattr(value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)


class ResourceProxyIter(TypedResourceIterable):
    """
    Returned by `ResourceProxy.proxy` if an iterable is supplied.
    """
    def __init__(self, objects, resource_type):
        # type: (Union[List[ResourceBase]], ResourceProxyBase) -> None
        super(ResourceProxyIter, self).__init__(resource_type)
        self.objects = objects

    def __iter__(self):
        resource_type = self.resource_type
        for obj in self.objects:
            yield resource_type.proxy(obj)


class ResourceProxyBase(ResourceBase):
    """
    Base proxy class
    """
    @classmethod
    def proxy(cls, resource):
        if hasattr(resource, '__iter__'):
            return ResourceProxyIter(resource, cls)
        elif isinstance(resource, ResourceBase):
            return cls(__resource=resource)
        raise TypeError("Only resource types can be proxied.")

    def __init__(self, *args, **kwargs):
        meta = getmeta(self)

        # Get shadowed resource if supplied
        shadow = kwargs.pop('__resource', None)
        if shadow is None:
            # Create a new instance
            self._shadow = meta.resource()
            super(ResourceProxyBase, self).__init__(*args, **kwargs)
        else:
            self._shadow = shadow

    def get_shadow(self):
        return self._shadow


class ResourceProxy(six.with_metaclass(ResourceProxyType, ResourceProxyBase)):
    """
    Proxy for a Resources that allow a filtered set of fields to be made
    available and updated.
    """



# -*- coding: utf-8 -*-
import re

# Typing imports
from typing import Iterable, Tuple, Union, Set, T  # noqa

_CAMEL_CASE_RE = re.compile(r'[A-Z]')
_LOWER_UNDERSCORE_CASE_RE = re.compile(r'_([a-z])')
_LOWER_DASH_CASE_RE = re.compile(r'-([a-z])')


def camel_to_lower_separated(s, sep):
    # type: (str, str) -> str
    """
    Convert camel case representation into lower separated case ie:

      backgroundColor -> background_color

    Note any separator at the start or end is stripped.

    """
    return _CAMEL_CASE_RE.sub(lambda m: sep + m.group(0).lower(), s).strip(sep)


def camel_to_lower_underscore(s):
    # type: (str) -> str
    """
    Convert camel case to lower underscore case.

        backgroundColor -> background_color

    """
    return camel_to_lower_separated(s, '_')


def camel_to_lower_dash(s):
    # type: (str) -> str
    """
    Convert camel case to lower dash case.

      backgroundColor -> background-color
    """
    return camel_to_lower_separated(s, '-')


def lower_underscore_to_camel(value):
    # type: (str) -> str
    """
    Convert lower underscore case to camel case

      background_color -> backgroundColor

    """
    return _LOWER_UNDERSCORE_CASE_RE.sub(
        lambda m: m.group(1).upper(),
        value.lower()
    )


def lower_dash_to_camel(value):
    # type: (str) -> str
    """
    Convert lower dash case to camel case

      background-color -> backgroundColor

    """
    return _LOWER_DASH_CASE_RE.sub(
        lambda m: m.group(1).upper(),
        value.lower()
    )


class cached_property(object):  # noqa - Made to match property builtin
    """
    Acts like a standard class `property` except return values cached.
    """
    @staticmethod
    def clear_caches(instance):
        instance._cache = {}

    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__
        self.__name__ = func.__name__
        self.__module__ = func.__module__

    def __get__(self, instance, owner):
        try:
            value = instance._cache[self.__name__]
        except (KeyError, AttributeError):
            value = self.func(instance)
            try:
                cache = instance._cache
            except AttributeError:
                cache = instance._cache = {}
            cache[self.__name__] = value
        return value


class lazy_property(object):  # noqa - Made to match the property builtin
    """
    The bottle cached property, requires a alternate name so as not to
    clash with existing cached_property behaviour
    """
    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = instance.__dict__[self.func.__name__] = self.func(instance)
        return value


EMPTY = []


def filter_fields(field_names, include=None, exclude=None, readonly=None):
    # type: (Iterable[str], Iterable[str], Iterable[str], Iterable[str]) -> Tuple[Set[str], Set[str]]
    """
    Filter a field iterable using the include/exclude/readonly options

    :param field_names: Iterable of field names as the source list
    :param include: Iterable of field names to be included
    :param exclude: Iterable of field names to be excluded
    :param readonly: Iterable of field names to be treated as read-only.
    :returns: A pair of sets of field names, the first is the set that have been selected,
        the second item is a set that have been selected and indicated to as read-only.

    """
    field_names = set(field_names)

    include = set(include or EMPTY)
    if include:
        field_names.intersection_update(include)

    exclude = set(exclude or EMPTY)
    if exclude:
        field_names.difference_update(exclude)

    readonly = set(readonly or EMPTY)
    if readonly:
        readonly.intersection_update(field_names)

    return field_names, readonly


def getmeta(resource_or_instance):
    """
    Get meta object from a resource or resource instance.

    :param resource_or_instance: Resource or instance of a resource.
    :type resource_or_instance: odin.resources.ResourceType | odin.resources.ResourceBase
    :return: Meta options class
    :rtype: odin.resources.ResourceOptions

    """
    return getattr(resource_or_instance, '_meta')


def field_iter(resource, include_virtual=True):
    """
    Return an iterator that yields fields from a resource.

    :param resource: Resource to iterate over.
    :param include_virtual: Include virtual fields.
    :returns: an iterator that returns fields.

    """
    meta = getmeta(resource)
    if include_virtual:
        return iter(meta.all_fields)
    else:
        return iter(meta.fields)


def field_iter_items(resource, fields=None):
    """
    Return an iterator that yields fields and their values from a resource.

    :param resource: Resource to iterate over.
    :param fields: Fields to use; if :const:`None` defaults to all of the resources fields.
    :returns: an iterator that returns (field, value) tuples.

    """
    meta = getmeta(resource)
    if fields is None:
        fields = meta.all_fields
    for f in fields:
        yield f, f.prepare(f.value_from_object(resource))


def virtual_field_iter_items(resource):
    """
    Return an iterator that yields virtual fields and their values from a resource.

    :param resource: Resource to iterate over.
    :returns: an iterator that returns (field, value) tuples.

    """
    meta = getmeta(resource)
    return field_iter_items(resource, meta.virtual_fields)


def attribute_field_iter_items(resource):
    """
    Return an iterator that yields fields and their values from a resource that have the attribute flag set.

    :param resource: Resource to iterate over.
    :returns: an iterator that returns (field, value) tuples.

    :note::
        This iterator is designed for codecs that have a distinction between attribute and element fields (eg XML).

    """
    return field_iter_items(resource, getmeta(resource).attribute_fields)


def element_field_iter_items(resource):
    """
    Return an iterator that yields fields and their values from a resource that do not have the attribute flag set.

    :param resource: Resource to iterate over.
    :returns: an iterator that returns (field, value) tuples.

    :note::
        This iterator is designed for codecs that have a distinction between attribute and element fields (eg XML).

    """
    return field_iter_items(resource, getmeta(resource).element_fields)


def extract_fields_from_dict(d, resource):
    """
    Extract values from a dict that are defined on a resource.

    Fields that are not found will not be included in the output dict.

    :param d: the source dictionary.
    :param resource: the resource that provides the fields.
    :returns: a dictionary of the resource fields that where found in the dict.

    """
    return {
        f.name: d[f.name] for f in field_iter(resource)
        if f.name in d
    }


def value_in_choices(value, choices):
    """
    Check if the value appears in the choices list (a iterable of tuples, the first value of which is the choice value)

    :param value:
    :param choices:
    :return: True if value is in the choices iterable.

    """
    for choice in choices:
        if value == choice[0]:
            return True
    return False


def iter_to_choices(i):
    # type: (Iterable[T]) -> tuple(Tuple[T, str])
    """
    Convert an iterator of strings (or types that can be converted to strings) and convert these into choice value
    pairs.
    """
    return tuple((v, str(v).title()) for v in i)


def force_tuple(value):
    # type: (Union[T, tuple(T), list(T)]) -> tuple(T)
    """
    Forces a value to be a tuple.

    Either by converting into a tuple (if is a list) or changing value to be a tuple.

    """
    if value is None:
        return ()
    if isinstance(value, tuple):
        return value
    if isinstance(value, list):
        return tuple(value)
    return value,


def chunk(iterable, n):
    # type: (Iterable[T], int) -> Iterable[Iterable[T]]
    """
    Return iterable of n items from an iterable.

    :param iterable: Iterable of items
    :param n: Size of iterable chunks to return.
    :return: Iterable chunk of input iterable

    """
    iterable = iter(iterable)
    state = {'continue': True}

    def inner():
        for _ in range(n):
            try:
                yield next(iterable)
            except StopIteration:
                state['continue'] = False

    while state['continue']:
        yield inner()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import datetime
import os

import pytest
from six import StringIO
from odin.codecs import toml_codec
from odin.exceptions import CodecDecodeError, ValidationError
from .resources import *

FIXTURE_PATH_ROOT = os.path.join(os.path.dirname(__file__), "fixtures")


class CustomString(object):
    def __init__(self, s):
        self.wrapped = s


class TestTomlCodec(object):
    def test_dumps_and_loads(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
        )

        data = toml_codec.dumps(in_resource)
        out_resource = toml_codec.loads(data)

        assert out_resource.title == in_resource.title
        assert out_resource.isbn == in_resource.isbn
        assert out_resource.num_pages == in_resource.num_pages
        assert out_resource.rrp == in_resource.rrp
        assert out_resource.fiction == in_resource.fiction
        assert out_resource.genre == in_resource.genre
        assert out_resource.authors[0].name == in_resource.authors[0].name
        assert out_resource.publisher.name == in_resource.publisher.name

    def test_dump_and_load(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
            published=[datetime.datetime(1987, 1, 1)],
        )

        fp = StringIO()
        toml_codec.dump(in_resource, fp)

        fp.seek(0)
        out_resource = toml_codec.load(fp)

        assert out_resource.title == in_resource.title
        assert out_resource.isbn == in_resource.isbn
        assert out_resource.num_pages == in_resource.num_pages
        assert out_resource.rrp == in_resource.rrp
        assert out_resource.fiction == in_resource.fiction
        assert out_resource.genre == in_resource.genre
        assert out_resource.authors[0].name == in_resource.authors[0].name
        assert out_resource.publisher.name == in_resource.publisher.name
        assert out_resource.published[0] == in_resource.published[0]

    def test_load__where_file_is_valid(self):
        with open(os.path.join(FIXTURE_PATH_ROOT, "book-valid.toml")) as f:
            data = toml_codec.load(f)

        assert data.books[0].isbn == "1-85723-394-8"

    def test_load__where_file_is_incorrect(self):
        """
        TOML is not formatted correctly
        """
        with pytest.raises(CodecDecodeError):
            with open(os.path.join(FIXTURE_PATH_ROOT, "book-incorrect.toml")) as f:
                toml_codec.load(f)

    def test_load__where_file_is_invalid(self):
        """
        TOML contains invalid field(s)
        """
        with pytest.raises(ValidationError):
            with open(os.path.join(FIXTURE_PATH_ROOT, "book-invalid.toml")) as f:
                toml_codec.load(f)

    def test_custom_type(self, monkeypatch):
        monkeypatch.setattr(toml_codec, "TOML_TYPES", {CustomString: lambda x: x.wrapped})

        in_resource = Book(
            title=CustomString("Consider Phlebas"),
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
            published=[datetime.datetime(1987, 1, 1)],
        )

        data = toml_codec.dumps(in_resource)
        out_resource = toml_codec.loads(data)

        assert out_resource.title == in_resource.title.wrapped



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import datetime
import os
from six import StringIO
from odin.codecs import yaml_codec
from .resources import *

FIXTURE_PATH_ROOT = os.path.join(os.path.dirname(__file__), "fixtures")


class TestYAMLCodec(object):
    def test_dumps_and_loads(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
        )

        data = yaml_codec.dumps(in_resource)
        out_resource = yaml_codec.loads(data)

        assert out_resource.title == in_resource.title
        assert out_resource.isbn == in_resource.isbn
        assert out_resource.num_pages == in_resource.num_pages
        assert out_resource.rrp == in_resource.rrp
        assert out_resource.fiction == in_resource.fiction
        assert out_resource.genre == in_resource.genre
        assert out_resource.authors[0].name == in_resource.authors[0].name
        assert out_resource.publisher.name == in_resource.publisher.name

    def test_dump_and_load(self):
        in_resource = Book(
            title="Consider Phlebas",
            isbn="0-333-45430-8",
            num_pages=471,
            rrp=19.50,
            fiction=True,
            genre="sci-fi",
            authors=[Author(name="Iain M. Banks")],
            publisher=Publisher(name="Macmillan"),
            published=[datetime.datetime(1987, 1, 1)],
        )

        fp = StringIO()
        yaml_codec.dump(in_resource, fp)

        fp.seek(0)
        out_resource = yaml_codec.load(fp)

        assert out_resource.title == in_resource.title
        assert out_resource.isbn == in_resource.isbn
        assert out_resource.num_pages == in_resource.num_pages
        assert out_resource.rrp == in_resource.rrp
        assert out_resource.fiction == in_resource.fiction
        assert out_resource.genre == in_resource.genre
        assert out_resource.authors[0].name == in_resource.authors[0].name
        assert out_resource.publisher.name == in_resource.publisher.name
        assert out_resource.published[0] == in_resource.published[0]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['odin',
 'odin.codecs',
 'odin.contrib',
 'odin.contrib.arrow',
 'odin.contrib.doc_gen',
 'odin.contrib.geo',
 'odin.contrib.inspect',
 'odin.contrib.money',
 'odin.contrib.pint',
 'odin.contrib.sphinx',
 'odin.contrib.swagger',
 'odin.fields',
 'odin.mapping',
 'odin.utils',
 'tests',
 'tests.contrib']

package_data = \
{'': ['*'],
 'odin.contrib.doc_gen': ['templates/odin/*'],
 'tests': ['fixtures/*']}

install_requires = \
['six']

extras_require = \
{':python_version < "3.4"': ['enum34'],
 ':python_version >= "2.7" and python_version < "2.8"': ['typing>=3.7.4,<3.8.0'],
 'arrow': ['arrow'],
 'filter_query': ['ply'],
 'inspect': ['humanfriendly'],
 'msgpack': ['msgpack'],
 'pint': ['pint'],
 'toml': ['toml'],
 'yaml': ['pyyaml']}

setup_kwargs = {
    'name': 'odin',
    'version': '1.6.1',
    'description': 'Data-structure definition/validation/traversal, mapping and serialisation toolkit for Python',
    'long_description': '\n####\nOdin\n####\n\nOdin provides a declarative framework for defining resources (classes) and their relationships, validation of the fields\nthat make up the resources and mapping between objects (either a resource, or other python structures).\n\nOdin also comes with built in serialisation tools for importing and exporting data from resources.\n\n+---------+-------------------------------------------------------------------------------------------------------------+\n| Docs/   | .. image:: https://readthedocs.org/projects/odin/badge/?version=latest                                      |\n| Help    |    :target: https://odin.readthedocs.org/                                                                   |\n|         |    :alt: ReadTheDocs                                                                                        |\n|         | .. image:: https://img.shields.io/badge/gitterim-timsavage.odin-brightgreen.svg?style=flat                  |\n|         |    :target: https://gitter.im/timsavage/odin                                                                |\n|         |    :alt: Gitter.im                                                                                          |\n+---------+-------------------------------------------------------------------------------------------------------------+\n| Build   | .. image:: https://github.com/python-odin/odin/actions/workflows/python-package.yml/badge.svg               |\n|         |    :target: https://github.com/python-odin/odin/actions/workflows/python-package.yml                        |\n|         |    :alt: Python package                                                                                     |\n|         | .. image:: https://api.dependabot.com/badges/status?host=github&repo=python-odin/odin                       |\n|         |    :target: https://dependabot.com                                                                          |\n|         |    :alt: Dependabot Status                                                                                  |\n+---------+-------------------------------------------------------------------------------------------------------------+\n| Quality | .. image:: https://sonarcloud.io/api/project_badges/measure?project=python-odin_odin&metric=sqale_rating    |\n|         |    :target: https://sonarcloud.io/dashboard?id=python-odin/odin                                             |\n|         |    :alt: Maintainability                                                                                    |\n|         | .. image:: https://sonarcloud.io/api/project_badges/measure?project=python-odin_odin&metric=security_rating |\n|         |    :target: https://sonarcloud.io/project/security_hotspots                                                 |\n|         |    :alt: Security                                                                                           |\n|         | .. image:: https://sonarcloud.io/api/project_badges/measure?project=python-odin_odin&metric=coverage        |\n|         |    :target: https://sonarcloud.io/code?id=python-odin_odin                                                  |\n|         |    :alt: Test Coverage                                                                                      |\n|         | .. image:: https://img.shields.io/badge/code%20style-black-000000.svg                                       |\n|         |    :target: https://github.com/ambv/black                                                                   |\n|         |    :alt: Once you go Black...                                                                               |\n+---------+-------------------------------------------------------------------------------------------------------------+\n| Package | .. image:: https://img.shields.io/pypi/v/odin                                                               |\n|         |    :target: https://pypi.io/pypi/odin/                                                                      |\n|         |    :alt: Latest Version                                                                                     |\n|         | .. image:: https://img.shields.io/pypi/pyversions/odin                                                      |\n|         |    :target: https://pypi.io/pypi/odin/                                                                      |\n|         | .. image:: https://img.shields.io/pypi/l/odin                                                               |\n|         |    :target: https://pypi.io/pypi/odin/                                                                      |\n|         | .. image:: https://img.shields.io/pypi/wheel/odin                                                           |\n|         |    :alt: PyPI - Wheel                                                                                       |\n|         |    :target: https://pypi.io/pypi/odin/                                                                      |\n+---------+-------------------------------------------------------------------------------------------------------------+\n\n\nHighlights\n**********\n\n* Class based declarative style\n* Fields for building composite resources\n* Field and Resource level validation\n* Easy extension to support custom fields\n* Python 2.7 :sup:`1`, Python 2.7+, Python 3.6+ and PyPy :sup:`1` supported\n* Support for documenting resources with `Sphinx <http://sphinx-doc.org/>`_\n* Minimal dependencies (base functionality only requires *six*)\n\n:sup:`1` certain contrib items are not supported. Pint is not installable with PyPy.\n\nUse cases\n*********\n* Design, document and validate complex (and simple!) data structures\n* Convert structures to and from different formats such as JSON, YAML, MsgPack, CSV, TOML\n* Validate API inputs\n* Define message formats for communications protocols, like an RPC\n* Map API requests to ORM objects\n\nQuick links\n***********\n\n* `Documentation <https://odin.readthedocs.org/>`_\n* `Project home <https://github.com/python-odin/odin>`_\n* `Issue tracker <https://github.com/python-odin/odin/issues>`_\n\n\nUpcoming features\n*****************\n\n**In development**\n\n* XML Codec (export completed)\n* Complete documentation coverage\n* Improvements for CSV Codec (writing, reading multi resource CSV\'s)\n* Integration with other libraries (ie `Django <https://www.djangoproject.com/>`_ Models/Forms)\n* Integration with SQLAlchemy\n\n\nRequires\n********\n\n* six\n\n**Optional**\n\n* simplejson - Odin will use simplejson if it is available or fallback to the builtin json library\n* msgpack-python - To enable use of the msgpack codec\n* pyyaml - To enable use of the YAML codec\n* toml - To enable use of the TOML codec\n\n**Contrib**\n\n* jinja2 >= 2.7 - For documentation generation\n* pint - Support for physical quantities using the `Pint <http://pint.readthedocs.org/>`_ library.\n\n**Development**\n\n* pytest - Testing\n* pytest-cov - Coverage reporting\n\nExample\n*******\n\n**With definition**::\n\n    import odin\n\n    class Author(odin.Resource):\n        name = odin.StringField()\n\n    class Publisher(odin.Resource):\n        name = odin.StringField()\n\n    class Book(odin.Resource):\n        title = odin.StringField()\n        authors = odin.ArrayOf(Author)\n        publisher = odin.DictAs(Publisher)\n        genre = odin.StringField()\n        num_pages = odin.IntegerField()\n\n::\n\n    >>> b = Book(\n            title="Consider Phlebas",\n            genre="Space Opera",\n            publisher=Publisher(name="Macmillan"),\n            num_pages=471\n        )\n    >>> b.authors.append(Author(name="Iain M. Banks"))\n    >>> from odin.codecs import json_codec\n    >>> json_codec.dumps(b, indent=4)\n    {\n        "$": "Book",\n        "authors": [\n            {\n                "$": "Author",\n                "name": "Iain M. Banks"\n            }\n        ],\n        "genre": "Space Opera",\n        "num_pages": 471,\n        "publisher": {\n            "$": "Publisher",\n            "name": "Macmillan"\n        },\n        "title": "Consider Phlebas"\n    }\n\n\nAuthors\n*******\n\nTim Savage\n\n\nSpecial Mention\n***************\n\nI would like to acknowledge the strong influence on the design of Odin Resources from the Django ORM and it\'s notable\ncontributor Malcolm Tredinnick. He was a valued colleague who\'s untimely passing left a large void in our company and\nthe wider community.\n',
    'author': 'Tim Savage',
    'author_email': 'tim@savage.company',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/python-odin/odin',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)
